import tensorflow_datasets as tfds
from os import mkdir
from os.path import join, exists
from nltk.tokenize import word_tokenize
import numpy as np

from a3t.dataset.download import download_wordvecs, download_artifacts

DATASET_HOME = join("/tmp", ".A3T")
if not exists(DATASET_HOME):
    mkdir(DATASET_HOME)


class Vocab:
    def __init__(self, g):
        self.str2id = g.str2id
        self.id2str = g.id2str

    def get_index(self, x):
        if x in self.str2id:
            return self.str2id[x]
        return self.str2id["_UNK_"]

    def get_word(self, x):
        return self.id2str[x]


class Glove:
    str2id = {"_UNK_": 0}
    id2str = ["_UNK_"]
    embedding = []
    is_built = False

    @staticmethod
    def build(all_vocab, dim, dataset_home=DATASET_HOME):
        if Glove.is_built:  # make sure it will only be built once
            return
        Glove.is_built = True
        glove_path = join(dataset_home, "glove")
        if not exists(glove_path):
            mkdir(glove_path)
        assert (all_vocab, dim) in [(6, 50), (6, 100), (6, 200), (6, 300), (42, 300), (840, 300)]
        glove_file = join(glove_path, "glove.%dB.%dd.txt" % (all_vocab, dim))
        cached_glove_npyfile = join(glove_path, "glove.%dB.%dd.npy" % (all_vocab, dim))
        cached_glove_dict_npyfile = join(glove_path, "glove_dict.%dB.%dd.npy" % (all_vocab, dim))
        cached_glove_file = join(glove_path, "glove.%dB.%dd.npy" % (all_vocab, dim))
        cached_glove_dict_file = join(glove_path, "glove_dict.%dB.%dd.npy" % (all_vocab, dim))

        if exists(cached_glove_npyfile) and exists(cached_glove_dict_file):
            Glove.embedding = np.load(cached_glove_npyfile)
            Glove.str2id = dict(np.load(cached_glove_dict_npyfile, allow_pickle=True).item())
            Glove.id2str = [None] * len(Glove.str2id)
            for x in Glove.str2id:
                Glove.id2str[Glove.str2id[x]] = x
            print("Loading cached glove embedding success!")
            return

        if not exists(glove_file):
            download_wordvecs(glove_path, all_vocab, dim)

        print("No cached glove embedding.")
        print("Using ", glove_file, " ...")

        lines = open(glove_file).readlines()

        for (i, line) in enumerate(lines):
            tmp = line.strip().split()
            Glove.embedding.append(np.array([float(x) for x in tmp[1:]]))
            Glove.str2id[tmp[0]] = i + 1
            Glove.id2str.append(tmp[0])

        # add embedding for UNK
        Glove.embedding = [np.zeros_like(Glove.embedding[0])] + Glove.embedding

        Glove.embedding = np.array(Glove.embedding)
        np.save(cached_glove_file, Glove.embedding)
        np.save(cached_glove_dict_file, Glove.str2id)
        print("Loading glove embedding success!")

    @staticmethod
    def make_vocab():
        assert Glove.is_built
        return Vocab(Glove)


class SST2CharLevel:
    train_X = []
    train_y = []
    test_X = []
    test_y = []
    val_X = []
    val_y = []
    max_len = 268
    is_built = False
    str2id = {"_UNK_": 0}
    id2str = ["_UNK_"]

    @staticmethod
    def map(c):
        if c not in SST2CharLevel.str2id:
            SST2CharLevel.str2id[c] = len(SST2CharLevel.str2id)
            SST2CharLevel.id2str.append(c)

        return SST2CharLevel.str2id[c]

    @staticmethod
    def build(dataset_home=DATASET_HOME):
        if SST2CharLevel.is_built:  # make sure it will only be built once
            return
        SST2CharLevel.is_built = True

        sst2char_path = join(dataset_home, "sst2char")
        if not exists(sst2char_path):
            mkdir(sst2char_path)
        download_artifacts(dataset_home)
        a3t_sst2test_file = join(dataset_home, "A3T-artifacts", "dataset", "sst2test.txt")

        def prepare_ds(ds):
            X = []
            y = []
            for features in tfds.as_numpy(ds):
                sentence, label = features["sentence"], features["label"]
                sentence = sentence.decode('UTF-8')
                x = np.zeros(SST2CharLevel.max_len, dtype=np.int)
                for (i, c) in enumerate(sentence):
                    x[i] = SST2CharLevel.map(c)
                X.append(x)
                y.append(label)

            return np.array(X), np.array(y)

        def prepare_test_ds(ds):
            X = []
            y = []
            for features in ds:
                features = features.strip()
                sentence, label = features[2:], features[:1]
                x = np.zeros(SST2CharLevel.max_len, dtype=np.int)
                for (i, c) in enumerate(sentence):
                    x[i] = SST2CharLevel.map(c)
                X.append(x)
                y.append(int(label))

            return np.array(X), np.array(y)

        ds_train = tfds.load(name="glue/sst2", split="train", shuffle_files=False)
        SST2CharLevel.train_X, SST2CharLevel.train_y = prepare_ds(ds_train)
        print("Loading training dataset success!")

        ds_val = tfds.load(name="glue/sst2", split="validation", shuffle_files=False)
        SST2CharLevel.val_X, SST2CharLevel.val_y = prepare_ds(ds_val)
        print("Loading validation dataset success!")

        SST2CharLevel.test_X, SST2CharLevel.test_y = prepare_test_ds(open(a3t_sst2test_file).readlines())
        print("Loading test dataset success!")

    @staticmethod
    def make_vocab():
        assert SST2CharLevel.is_built
        return Vocab(SST2CharLevel)


class SST2WordLevel:
    train_X = []
    train_y = []
    test_X = []
    test_y = []
    val_X = []
    val_y = []
    max_len = 56
    is_built = False
    vocab = None

    @staticmethod
    def build(dataset_home=DATASET_HOME):
        if SST2WordLevel.is_built:  # make sure it will only be built once
            return
        SST2WordLevel.is_built = True
        SST2WordLevel.vocab = Glove.make_vocab()

        sst2word_path = join(dataset_home, "sst2word")
        if not exists(sst2word_path):
            mkdir(sst2word_path)
        download_artifacts(dataset_home)
        a3t_sst2test_file = join(dataset_home, "A3T-artifacts", "dataset", "sst2test.txt")

        def prepare_ds(ds):
            X = []
            y = []
            for features in tfds.as_numpy(ds):
                sentence, label = features["sentence"], features["label"]
                tokens = word_tokenize(sentence.decode('UTF-8'))
                x = np.zeros(SST2WordLevel.max_len, dtype=np.int)
                for (i, token) in enumerate(tokens):
                    x[i] = SST2WordLevel.vocab.get_index(token)
                X.append(x)
                y.append(label)

            return np.array(X), np.array(y)

        def prepare_test_ds(ds):
            X = []
            y = []
            for features in ds:
                features = features.strip()
                sentence, label = features[2:], features[:1]
                tokens = word_tokenize(sentence)
                x = np.zeros(SST2WordLevel.max_len, dtype=np.int)
                for (i, token) in enumerate(tokens):
                    x[i] = SST2WordLevel.vocab.get_index(token)
                X.append(x)
                y.append(int(label))

            return np.array(X), np.array(y)

        ds_train = tfds.load(name="glue/sst2", split="train", shuffle_files=False)
        SST2WordLevel.train_X, SST2WordLevel.train_y = prepare_ds(ds_train)
        print("Loading training dataset success!")

        ds_val = tfds.load(name="glue/sst2", split="validation", shuffle_files=False)
        SST2WordLevel.val_X, SST2WordLevel.val_y = prepare_ds(ds_val)
        print("Loading validation dataset success!")

        SST2WordLevel.test_X, SST2WordLevel.test_y = prepare_test_ds(open(a3t_sst2test_file).readlines())
        print("Loading test dataset success!")



import numpy as np
import unittest
import random
import itertools

from a3t.DSL.general_HotFlip import GeneralHotFlipAttack, Candidate
from a3t.DSL.transformation import Sub, Ins, Del
from a3t.utils import Beam


class TestModel1:
    def __init__(self, dim):
        self.embed = {}
        self.dim = dim


class TestModel:
    def __init__(self, dim):
        self.embed = {}
        self.dim = dim

    def get_grad(self, x, y):
        self.scan(x)
        return np.array([self.embed[w] for w in x])

    def get_embed(self, x, ret_len=None):
        self.scan(x)
        if ret_len is None:
            ret_len = len(x)
        ret = np.zeros((ret_len, self.dim))
        for (i, w) in enumerate(x):
            ret[i] = self.embed[w]
        return ret

    def scan(self, x):
        for w in x:
            if w not in self.embed:
                self.embed[w] = np.random.random(self.dim)


class TestGeneralHotFlipAttack(unittest.TestCase):
    def test_arguments(self):
        with self.assertRaises(AttributeError):
            GeneralHotFlipAttack([(1, 3, 4)])
        with self.assertRaises(AttributeError):
            GeneralHotFlipAttack([(1, 2)])
        attack = GeneralHotFlipAttack([(Sub("dataset", True), 1), (Del(), 2), (Ins(), 3)])
        with self.assertRaises(AttributeError):
            attack.gen_adv(TestModel1(100), [], 1, 5, None)

    def test_beam(self):
        beams = [Beam(4), Beam(3), Beam(2)]
        test_sets = [[("aaa", 1), ("bbb", 0), ("ccc", -3)], [("aaa", 1), ("bbb", -3), ("ccc", 0)],
                     [("aaa", 1), ("bbb", -3), ("ccc", 0), ("ddd", 2)]]
        for (candidates, tests) in zip(beams, test_sets):
            random.shuffle(tests)
            for a in tests:
                candidates.add(*a)
            tests.sort(key=lambda x: -x[1])
            tests = tests[:candidates.budget]
            ans = candidates.check_balance()
            self.assertEqual(len(tests), len(ans))
            for (a, b) in zip(tests, ans):
                self.assertTupleEqual(a, b)

    def test_perturbation(self):
        model = TestModel(100)
        perturb = [(Sub("dataset", True), 3), (Del(), 2), (Ins(), 1)]
        attack = GeneralHotFlipAttack(perturb)
        sen = "i see that a cat sits on the floor .".split()
        sub_pos = perturb[0][0].get_pos(sen)
        del_pos = perturb[1][0].get_pos(sen)
        ins_pos = perturb[2][0].get_pos(sen)
        cnt = 0
        for sub_cnt in range(perturb[0][1] + 1):
            for del_cnt in range(perturb[1][1] + 1):
                for ins_cnt in range(perturb[2][1] + 1):
                    for a in itertools.permutations(sub_pos, sub_cnt):
                        for b in itertools.permutations(del_pos, del_cnt):
                            for c in itertools.permutations(ins_pos, ins_cnt):
                                a = set(a)
                                b = set(b)
                                c = set(c)
                                if a.isdisjoint(b) and a.isdisjoint(c) and b.isdisjoint(c):
                                    cnt += 1

        ans = attack.gen_adv(model, sen, 0, 10000, model.get_embed, True)
        self.assertEqual(cnt, len(ans))
        for i in range(1, len(ans)):
            self.assertTrue(ans[i][1] <= ans[i - 1][1])

    def test_multi_choice(self):
        model = TestModel(100)
        perturb = [(Sub("dataset", False), 3)]
        attack = GeneralHotFlipAttack(perturb)
        sen = "i see that a cat sits on the floor .".split()
        sub_pos = perturb[0][0].get_pos(sen)
        for i, (start_pos, end_pos) in enumerate(sub_pos):
            cnt = 0
            for _ in perturb[0][0].transformer(sen, start_pos, end_pos):
                cnt += 1
            sub_pos[i] = cnt

        cnt = 0
        for sub_cnt in range(perturb[0][1] + 1):
            for a in itertools.permutations(sub_pos, sub_cnt):
                cnt += np.prod(list(a))

        ans = attack.gen_adv(model, sen, 0, 10000, model.get_embed, True)
        self.assertEqual(cnt, len(ans))
        for i in range(1, len(ans)):
            self.assertTrue(ans[i][1] <= ans[i - 1][1])

    def test_single_transformation(self):
        model = TestModel(100)
        perturb1 = [(Sub("dataset", True), 2)]
        perturb2 = [(Del(), 2)]
        perturb3 = [(Ins(), 2)]
        attack1 = GeneralHotFlipAttack(perturb1)
        attack2 = GeneralHotFlipAttack(perturb2)
        attack3 = GeneralHotFlipAttack(perturb3)
        sen = "i see that a cat sits on the floor .".split()
        sub_pos = perturb1[0][0].get_pos(sen)
        exp_ans1 = []
        for sub_cnt in range(perturb1[0][1] + 1):
            for a in itertools.permutations(sub_pos, sub_cnt):
                new_x = sen
                for (start_pos, end_pos) in a:
                    new_x = next(perturb1[0][0].transformer(new_x, start_pos, end_pos))
                exp_ans1.append(new_x)

        del_pos = perturb2[0][0].get_pos(sen)
        exp_ans2 = []
        for del_cnt in range(perturb2[0][1] + 1):
            for a in itertools.permutations(del_pos, del_cnt):
                new_x = []
                for (i, w) in enumerate(sen):
                    if (i, i + 1) not in a:
                        new_x.append(w)

                exp_ans2.append(new_x)

        ins_pos = perturb3[0][0].get_pos(sen)
        exp_ans3 = []
        for ins_cnt in range(perturb3[0][1] + 1):
            for a in itertools.permutations(ins_pos, ins_cnt):
                new_x = []
                for (i, w) in enumerate(sen):
                    if (i, i + 1) not in a:
                        new_x.append(w)
                    else:
                        new_x.append(w)
                        new_x.append(w)

                exp_ans3.append(new_x)

        ans1 = attack1.gen_adv(model, sen, 0, 10000, model.get_embed)
        ans2 = attack2.gen_adv(model, sen, 0, 10000, model.get_embed)
        ans3 = attack3.gen_adv(model, sen, 0, 10000, model.get_embed)

        def list2map(l):
            ret = {}
            for x in l:
                x = tuple(x)
                ret[x] = ret.setdefault(x, 0) + 1
            return ret

        self.assertDictEqual(list2map(exp_ans1), list2map(ans1))
        self.assertDictEqual(list2map(exp_ans2), list2map(ans2))
        self.assertDictEqual(list2map(exp_ans3), list2map(ans3))

    def test_map_ori2x(self):
        model = TestModel(100)
        perturb = [(Sub("dataset", True), 3), (Del({"floor"}), 2), (Ins(), 1)]
        sen = "i see that a cat sits on the floor .".split()
        sub_pos = perturb[0][0].get_pos(sen)
        del_pos = perturb[1][0].get_pos(sen)
        ins_pos = perturb[2][0].get_pos(sen)
        candidate = Candidate(["dummy"], 0, [None] * len(sen))
        candidates = Beam(1)
        candidate.try_all_pos(ins_pos, perturb[2][0], model.get_grad(candidate.x, 0), model.get_embed,
                              candidates)
        self.assertEqual(0, len(candidates.queue))
        map_ori2x = [None] * len(sen)
        map_ori2x[sen.index("floor")] = 1
        candidate = Candidate(["dummy1", "floor", "dummy2"], 0, map_ori2x)  # all perturbation can be applied
        candidates = Beam(1)
        candidate.try_all_pos(sub_pos, perturb[0][0], model.get_grad(candidate.x, 0), model.get_embed,
                              candidates)
        self.assertSequenceEqual(["dummy1", "flooring", "dummy2"], candidates.check_balance()[0][0].x)

        candidates = Beam(1)
        candidate.try_all_pos(del_pos, perturb[1][0], model.get_grad(candidate.x, 0), model.get_embed,
                              candidates)
        self.assertSequenceEqual(["dummy1", "dummy2"], candidates.check_balance()[0][0].x)

        candidates = Beam(1)
        candidate.try_all_pos(ins_pos, perturb[2][0], model.get_grad(candidate.x, 0), model.get_embed,
                              candidates)
        self.assertSequenceEqual(["dummy1", "floor", "floor", "dummy2"], candidates.check_balance()[0][0].x)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import heapq

inf = 1e10


def tuple_set_union(ret0: tuple, ret1: tuple):
    if ret0 is None:
        return ret1
    if ret1 is None:
        return ret0
    ret = ()
    max_len = max(len(ret0), len(ret1))
    for i in range(max_len):
        if i >= len(ret0):
            r0 = [""]
        else:
            r0 = ret0[i]
        if i >= len(ret1):
            r1 = [""]
        else:
            r1 = ret1[i]
        ret += (tuple(set(r0).union(set(r1))),)

    return ret


class Beam:
    def __init__(self, budget):
        '''
        The queue are contain two elements: (data, score), while the score is the ranking key (descending)
        :param budget: the beam search budget
        '''
        self.budget = budget
        self.queue = []
        self.in_queue = {}

    def add(self, data, score):
        '''
        add (data, score) into queue
        :param data: candidate
        :param score: score of the candidate
        :return: True if added in, False otherwise
        '''
        if data in self.in_queue:  # if data is ready in the priority queue, we update the score in self.in_queue
            if self.in_queue[data] < score:
                self.in_queue[data] = score
                return True
            return False

        ret = True
        if len(self.queue) == self.budget:  # if size(queue) == budget, we need to remove one
            while True:
                a, b = heapq.heappop(self.queue)
                # the top of the priority queue may not be smallest, because it may contain new value in self.in_queue
                if a == self.in_queue[b]:  # if the value is not updated, then it is smallest
                    break  # remove (a, b)
                else:
                    heapq.heappush(self.queue,
                                   (self.in_queue[b], b))  # otherwise, update in the priority queue (lazily)
            del self.in_queue[b]  # remove (a, b) from self.in_queue
            if a > score:  # if the old (a, b) is better then new (score, data), we replace (score, data) with (a, b)
                score, data = a, b
                ret = False

        heapq.heappush(self.queue, (score, data))  # add (score, data)
        self.in_queue[data] = score  # update in self.in_queue
        return ret

    def extend(self, others):
        if isinstance(others, list):
            for data, score in others:
                self.add(data, score)
        else:
            assert False
            # for data, score in others.queue:
            #     self.add(data, score)

    def check_balance(self):
        ret = []
        for data in self.in_queue:
            ret.append((data, self.in_queue[data]))
        ret.sort(key=lambda x: -x[1])
        return ret

    def is_same(self, others: list):
        if len(others) != len(self.queue):
            return False
        others.sort(key=lambda x: -x[1])
        for i in range(len(others)):
            data, score = others[i]
            if data not in self.in_queue or self.in_queue[data] != score:
                return False

        return True


class UnorderedBeam:
    def __init__(self, budget):
        self.budget = budget
        self.queue = []

    def add(self, data):
        self.queue.append(data)

    def extend(self, others):
        if isinstance(others, list):
            self.queue.extend(others)
        else:
            assert False
            # for data, score in others.queue:
            #     self.add(data, score)

    def check_balance(self):
        ids = np.random.randint(0, len(self.queue), self.budget)
        ret = []
        for id in ids:
            ret.append(self.queue[id])
        return ret


def swap_pytorch(x, p1, p2):
    z = x[p1].clone()
    x[p1] = x[p2]
    x[p2] = z



from abc import ABC, abstractmethod
import os
from nltk import pos_tag
import gzip

from a3t.dataset.download import download_ppdb, download_artifacts


class Transformation(ABC):
    def __init__(self, length_preserving=False):
        """
        A default init function.
        """
        super().__init__()
        self.length_preserving = length_preserving

    @abstractmethod
    def get_pos(self, ipt):
        # get matched positions in input
        pass

    @abstractmethod
    def transformer(self, ipt, start_pos, end_pos):
        # transformer for a segment of input
        pass

    def sub_transformer(self, ipt, start_pos, end_pos):
        # substring transformer for length preserving transformation
        assert self.length_preserving


class Sub(Transformation):
    def __init__(self, use_fewer_sub, dataset_home="/tmp/.A3T", ppdb_type="s"):
        """
        Init Sub transformation
        Sub substitutes one word with its synonym with a requirement that they have the same pos_tag.
        :param use_fewer_sub: if True, one word can only be substituted by one of its synonyms.
        :param dataset_home: the home of the dataset
        :param ppdb_type the type of the ppdb synonyms can be ["s", "m", "l", "xl", "xxl", "xxxl"]
        """
        assert ppdb_type in ["s", "m", "l", "xl", "xxl", "xxxl"]
        self.use_fewer_sub = use_fewer_sub
        self.synonym_dict = {}
        self.synonym_dict_pos_tag = {}
        pddb_path = os.path.join(dataset_home, "ppdb")
        if not os.path.exists(pddb_path):
            os.mkdir(pddb_path)
        pddb_file = os.path.join(pddb_path, "ppdb-2.0-%s-lexical" % ppdb_type)
        if not os.path.exists(pddb_file):
            download_ppdb(pddb_path, ppdb_type)

        with gzip.open(pddb_file, 'rb') as f:
            lines = f.readlines()
            for line in lines:
                line = line.decode()
                tmp = line.strip().split(" ||| ")
                postag, x, y = tmp[0][1:-1], tmp[1], tmp[2]
                self.synonym_dict_add_str(x, y, postag)
                self.synonym_dict_add_str(y, x, postag)

        print("Compute synonym_dict success!")
        super(Sub, self).__init__(True)

    def synonym_dict_add_str(self, x, y, postag):
        if x not in self.synonym_dict:
            self.synonym_dict[x] = [y]
            self.synonym_dict_pos_tag[x] = [postag]
        elif not self.use_fewer_sub:
            self.synonym_dict[x].append(y)
            self.synonym_dict_pos_tag[x].append(postag)

    def get_pos(self, ipt):
        ipt_pos_tag = pos_tag(ipt)
        ret = []
        for (start_pos, (x, pos_tagging)) in enumerate(ipt_pos_tag):
            if x in self.synonym_dict:
                if any(pos_tagging == t for t in self.synonym_dict_pos_tag[x]):
                    ret.append((start_pos, start_pos + 1))

        return ret

    def transformer(self, ipt, start_pos, end_pos):
        ipt_pos_tag = pos_tag(ipt)
        x = ipt[start_pos]
        pos_tagging = ipt_pos_tag[start_pos][1]
        for (w, t) in zip(self.synonym_dict[x], self.synonym_dict_pos_tag[x]):
            if t == pos_tagging:
                new_ipt = ipt[:start_pos] + [w] + ipt[end_pos:]
                yield new_ipt

    def sub_transformer(self, ipt, start_pos, end_pos):
        ipt_pos_tag = pos_tag(ipt)
        x = ipt[start_pos]
        pos_tagging = ipt_pos_tag[start_pos][1]
        for (w, t) in zip(self.synonym_dict[x], self.synonym_dict_pos_tag[x]):
            if t == pos_tagging:
                yield [w]


class SubChar(Transformation):
    def __init__(self, use_fewer_sub, dataset_home="/tmp/.A3T"):
        """
        Init SubChar transformation
        Sub substitutes one char with other chars. The substitution relation is one direction (x => y, but not x <= y)
        :param sub_file: the character substitution file
        :param use_fewer_sub: if True, one char can only be substituted by one of its synonyms.
        """
        download_artifacts(dataset_home)
        a3t_sst2test_file = os.path.join(dataset_home, "A3T-artifacts", "dataset", "en.key")
        lines = open(a3t_sst2test_file).readlines()
        self.synonym_dict = {}
        for line in lines:
            tmp = line.strip().split()
            x, y = tmp[0], tmp[1]
            if x not in self.synonym_dict:
                self.synonym_dict[x] = [y]
            elif not use_fewer_sub:
                self.synonym_dict[x].append(y)

        print("Compute set success!")
        super(SubChar, self).__init__(True)

    def get_pos(self, ipt):
        ret = []
        for (start_pos, x) in enumerate(ipt):
            if x in self.synonym_dict:
                ret.append((start_pos, start_pos + 1))

        return ret

    def transformer(self, ipt, start_pos, end_pos):
        x = ipt[start_pos]
        for w in self.synonym_dict[x]:
            new_ipt = ipt[:start_pos] + [w] + ipt[end_pos:]
            yield new_ipt

    def sub_transformer(self, ipt, start_pos, end_pos):
        x = ipt[start_pos]
        for w in self.synonym_dict[x]:
            yield [w]


class Del(Transformation):
    def __init__(self, stop_words=None):
        """
        Init Del transformation
        Del removes one stop word.
        :param stop_words: stop words to delete
        """
        if stop_words is None:
            stop_words = {"a", "and", "the", "of", "to"}
        self.stop_words = stop_words

        super(Del, self).__init__()

    def get_pos(self, ipt):
        ret = []
        for (start_pos, x) in enumerate(ipt):
            if x in self.stop_words:
                ret.append((start_pos, start_pos + 1))

        return ret

    def transformer(self, ipt, start_pos, end_pos):
        new_ipt = ipt[:start_pos] + ipt[end_pos:]
        yield new_ipt


class Ins(Transformation):
    def __init__(self):
        """
        Init Ins transformation
        Ins duplicates a word behind it.
        """
        super(Ins, self).__init__()

    def get_pos(self, ipt):
        return [(x, x + 1) for x in range(len(ipt))]

    def transformer(self, ipt, start_pos, end_pos):
        new_ipt = ipt[:start_pos] + [ipt[start_pos], ipt[start_pos]] + ipt[end_pos:]
        yield new_ipt


class InsChar(Transformation):
    def __init__(self, use_fewer_sub, dataset_home="/tmp/.A3T"):
        """
        Init SubChar transformation
        Sub substitutes one char with other chars. The substitution relation is one direction (x => y, but not x <= y)
        :param sub_file: the character substitution file
        :param use_fewer_sub: if True, one char can only be substituted by one of its synonyms.
        """
        download_artifacts(dataset_home)
        a3t_sst2test_file = os.path.join(dataset_home, "A3T-artifacts", "dataset", "en.key")
        lines = open(a3t_sst2test_file).readlines()
        self.synonym_dict = {}
        for line in lines:
            tmp = line.strip().split()
            x, y = tmp[0], tmp[1]
            if x not in self.synonym_dict:
                self.synonym_dict[x] = [y]
            elif not use_fewer_sub:
                self.synonym_dict[x].append(y)

        print("Compute set success!")
        super(InsChar, self).__init__()

    def get_pos(self, ipt):
        ret = []
        for (start_pos, x) in enumerate(ipt):
            if x in self.synonym_dict:
                ret.append((start_pos, start_pos + 1))

        return ret

    def transformer(self, ipt, start_pos, end_pos):
        x = ipt[start_pos]
        for w in self.synonym_dict[x]:
            new_ipt = ipt[:start_pos] + [x, w] + ipt[end_pos:]
            yield new_ipt


class Swap(Transformation):
    def __init__(self):
        """
        Init Swap transformation
        Swap two adjacent tokens.
        """
        super(Swap, self).__init__(True)

    def get_pos(self, ipt):
        return [(x, x + 2) for x in range(len(ipt) - 1)]

    def transformer(self, ipt, start_pos, end_pos):
        new_ipt = ipt[:start_pos] + [ipt[start_pos + 1], ipt[start_pos]] + ipt[end_pos:]
        yield new_ipt

    def sub_transformer(self, ipt, start_pos, end_pos):
        yield [ipt[start_pos + 1], ipt[start_pos]]



import numpy as np

from a3t.DSL.transformation import Transformation
from a3t.utils import Beam


class GeneralHotFlipAttack:
    def __init__(self, perturbation: list, use_random_aug=False):
        """
        :param list perturbation: A perturbation space specified in the DSL.
        For example, [(Sub(), 2), (Del(), 1)] means at most 2 Sub string transformations and at most 1 Del string
        transformation. Sub and Del are default string transformations (see transformation.py).

        :Classifier Capacity: Gradient

        The idea of writing a perturbation space in a DSL is first proposed and implemented in the following paper:
        Robustness to Programmable String Transformations via Augmented Abstract Training.
        Yuhao Zhang, Aws Albarghouthi, Loris D’Antoni.
        `[pdf] <https://arxiv.org/abs/2002.09579>`__
        `[code] <https://github.com/ForeverZyh/A3T>`__

        """

        try:
            for (a, b) in perturbation:
                assert isinstance(a, Transformation)
                assert isinstance(b, int)
        except:
            raise AttributeError("param perturbation %s is not in the correct form." % str(perturbation))

        self.perturbation = perturbation
        self.use_random_aug = use_random_aug

    def gen_adv(self, model, x: list, y, top_n: int, get_embed, return_score=False):
        """
        Beam search for the perturbation space. The order of beam search is the same in the perturbation DSL.
        Some adversarial attack tries to rearrange the order of beam search for better effectiveness.
        TODO: the order of beam search can be rearranged for better effectiveness.
        :param model: the victim model, which has to support a method get_grad.
        :param x: a list of input tokens.
        :param y: the correct label of input x.
        :param top_n: maximum number of adversarial candidates given the perturbation space.
        :param get_embed: get_embed(x, ret_len) takes a list of tokens as inputs and output the embedding matrix with
        shape (ret_len, dim).
        if ret_len > len(x), then padding is needed.
        if ret_len < len(x), then truncating is needed. (Currently, we do not need truncate)
        :param return_score: whether return the score as a list [(sen, score)], default False, i.e., return [sen]
        :return: a list of adversarial examples
        """

        try:
            model.get_grad
        except AttributeError:
            raise AttributeError("The victim model does not support get_grad method.")

        candidate = Candidate(x, 0 if not self.use_random_aug else np.random.random())
        candidates = Beam(top_n)
        candidates.add(candidate, candidate.score)
        for (tran, delta) in self.perturbation:
            possible_pos = tran.get_pos(x)  # get a list of possible positions
            perturbed_set = set()  # restore the perturbed candidates to eliminate duplications
            for _ in range(delta):
                # keep the old candidates because we will change candidates in the following loop
                old_candidates = candidates.check_balance()
                for (candidate, _) in old_candidates:
                    if candidate not in perturbed_set:
                        if len(candidate.x) > 0:
                            if self.use_random_aug:
                                candidate.try_all_pos(possible_pos, tran, None, None, candidates)
                            else:
                                candidate.try_all_pos(possible_pos, tran, model.get_grad(candidate.x, y), get_embed,
                                                      candidates)
                        perturbed_set.add(candidate)

        ret = candidates.check_balance()
        if return_score:
            return [(x.x, x.score) for (x, _) in ret]
        else:
            return [x.x for (x, _) in ret]


class Candidate:
    def __init__(self, x, score, map_ori2x=None):
        """
        Init a candidate
        :param x: input tokens
        :param score: score of adversarial attack, the larger the better
        :param map_ori2x: position mapping from the original input to transformed input
        """
        self.x = x
        self.score = score
        if map_ori2x is None:
            self.map_ori2x = [_ for _ in range(len(x))]
        else:
            self.map_ori2x = map_ori2x

    def __lt__(self, other):
        return self.x < other.x

    def try_all_pos(self, pos: list, tran: Transformation, gradients, get_embed, candidates: Beam):
        """
        Try all possible positions for trans
        :param pos: possible positions, a list of (start_pos, end_pos)
        :param tran: the target transformation
        :param gradients: the gradients tensor with respect to self.x
        :param get_embed: a function for getting the embedding of a list of tokens
        :param candidates: a beam of candidates, will be modified by this methods
        :return: None
        """

        for (start_pos_ori, end_pos_ori) in pos:
            if all(self.map_ori2x[i] is not None for i in range(start_pos_ori, end_pos_ori)):
                start_pos_x = self.map_ori2x[start_pos_ori]
                # notice that self.map_ori2x[end_pos] can be None, we need to calculate from self.map_ori2x[end_pos - 1]
                end_pos_x = self.map_ori2x[end_pos_ori - 1] + 1

                for new_x in tran.transformer(self.x, start_pos_x, end_pos_x):
                    delta_len = len(new_x) - len(self.x)
                    if get_embed is not None and gradients is not None:
                        old_embedding = get_embed(self.x[start_pos_x:])
                        # ret_len specifies the ret length (padding if not enought)
                        new_embedding = get_embed(new_x[start_pos_x:min(len(new_x), len(self.x))],
                                                  ret_len=len(self.x) - start_pos_x)
                        # gradients[start_pos_x:] has shape (len(self.x) - start_pos_x, dim)
                        new_score = self.score + np.sum(gradients[start_pos_x:] * (new_embedding - old_embedding))
                    else:  # we use random sampling
                        new_score = np.random.random()
                    new_map_ori2x = self.map_ori2x[:start_pos_ori] + [None] * (end_pos_ori - start_pos_ori) + [
                        p if p is None else p + delta_len for p in self.map_ori2x[end_pos_ori:]]
                    new_candidate = Candidate(new_x, new_score, new_map_ori2x)
                    candidates.add(new_candidate, new_score)



from a3t.diffai.train import train
from a3t.dataset.dataset_loader import SST2WordLevel, SST2CharLevel, Glove
from a3t.diffai.helpers import loadDataset
from a3t.DSL.transformation import Sub, SubChar, Del, Ins, InsChar, Swap


def prepare_SST2WordLevel():
    Glove.build(6, 50)
    SST2WordLevel.build()


def prepare_SST2CharLevel():
    SST2CharLevel.build()


class Args:
    batch_size = 10
    model_srt = ""
    clip_norm = False
    epochs = 20
    log_freq = 10
    save_freq = 1
    lr = 1e-3
    threshold = -0.01
    lr_patience = 0
    early_stop_patience = 5
    factor = 0.5
    max_norm = 10000
    width = 0.01
    out = "out"
    adv_train_num = 2
    adv_test_num = 0
    epoch_perct_decay = 0.6
    train_lambda = 0.75
    regularize = 0.005
    seed = 12345


# load data

prepare_SST2WordLevel()
sst2word_loader_train = loadDataset(SST2WordLevel, Args.batch_size, "train", test_slice=slice(0, 3000, None))
sst2word_loader_val = loadDataset(SST2WordLevel, Args.batch_size, "val", test_slice=slice(0, 100, None))
sst2word_loader_test = loadDataset(SST2WordLevel, Args.batch_size, "test", test_slice=slice(0, 100, None))

prepare_SST2CharLevel()
sst2char_loader_train = loadDataset(SST2CharLevel, Args.batch_size, "train", test_slice=slice(0, 30, None))
sst2char_loader_val = loadDataset(SST2CharLevel, Args.batch_size, "val", test_slice=slice(0, 10, None))
sst2char_loader_test = loadDataset(SST2CharLevel, Args.batch_size, "test", test_slice=slice(0, 10, None))

# transformations
word_perturbation = [(Sub(True), 2), (Ins(), 2), (Del(), 2)]
char_perturbation = [(SubChar(True), 2), (InsChar(True), 2), (Del(set()), 2), (Swap(), 2)]
# print(word_perturbation[0][0].synonym_dict)
# print(char_perturbation[0][0].synonym_dict)

Args.model_srt = "CharLevelSST2"
# Train
train(SST2CharLevel.make_vocab(), sst2char_loader_train, sst2char_loader_val, sst2char_loader_test,
      [(InsChar(True), 2), (Del(set()), 2)], [(SubChar(True), 2), (Swap(), 2)], fixed_len=SST2CharLevel.max_len,
      args=Args)

# # Test
# train(SST2CharLevel.make_vocab(), sst2char_loader_train, sst2char_loader_val, sst2char_loader_test,
#       [(InsChar(True), 2), (Del(set()), 2)], [(SubChar(True), 2), (Swap(), 2)], fixed_len=SST2CharLevel.max_len,
#       args=Args, test=True,
#       load_path="out/XXX.pynet")


Args.model_srt = "WordLevelSST2"
# Train
train(SST2WordLevel.vocab, sst2word_loader_train, sst2word_loader_val, sst2word_loader_test,
      [(Ins(), 2), (Del(), 2)], [(Sub(True), 2)], fixed_len=SST2WordLevel.max_len, args=Args)

# # Test
# train(SST2WordLevel.vocab, sst2word_loader_train, sst2word_loader_val, sst2word_loader_test,
#       [(Ins(), 2), (Del(), 2)], [(Sub(True), 2)], fixed_len=SST2WordLevel.max_len, args=Args,
#       test=True,
#       load_path="out/YYY.pynet")



"""
Modified from https://github.com/stanfordnlp/treelstm/blob/master/scripts/download.py
"""
from urllib.request import urlopen
import sys
import os
import zipfile


def download(url, dirpath):
    filename = url.split('/')[-1]
    filepath = os.path.join(dirpath, filename)
    u = urlopen(url)
    try:
        f = open(filepath, 'wb')
    except:
        print("Cannot write %s" % filepath)
        raise Exception
    try:
        filesize = int(u.info().get_all("Content-Length")[0])
    except:
        print("URL %s failed to report length" % url)
        filesize = None
    print("Downloading: %s Bytes: %s" % (filename, filesize))

    downloaded = 0
    block_sz = 8192
    status_width = 70
    while True:
        buf = u.read(block_sz)
        if not buf:
            print('')
            break
        else:
            print('', end='\r')
        downloaded += len(buf)
        f.write(buf)
        if filesize is not None:
            status = (("[%-" + str(status_width + 1) + "s] %3.2f%%") %
                      ('=' * int(float(downloaded) / filesize * status_width) + '>', downloaded * 100. / filesize))
            print(status, end='')
        sys.stdout.flush()
    f.close()
    return filepath


def unzip(filepath):
    print("Extracting: " + filepath)
    dirpath = os.path.dirname(filepath)
    with zipfile.ZipFile(filepath) as zf:
        zf.extractall(dirpath)
    os.remove(filepath)


def download_wordvecs(dirpath, all_vocab, dim):
    if all_vocab == 6:
        url = 'http://www-nlp.stanford.edu/data/glove.6B.zip'
    else:
        url = 'http://www-nlp.stanford.edu/data/glove.%dB.%dd.zip' % (all_vocab, dim)
    unzip(download(url, dirpath))


def download_ppdb(dirpath, ppdb_type):
    download("http://nlpgrid.seas.upenn.edu/PPDB/eng/ppdb-2.0-%s-lexical" % ppdb_type, dirpath)


def download_artifacts(dataset_home):
    a3t_artifact_file = os.path.join(dataset_home, "A3T-artifacts", "dataset")
    a3t_enkey_file = os.path.join(a3t_artifact_file, "en.key")
    a3t_sst2test_file = os.path.join(a3t_artifact_file, "sst2test.txt")
    if not os.path.exists(a3t_enkey_file) or not os.path.exists(a3t_sst2test_file):
        unzip(download("https://github.com/ForeverZyh/A3T/archive/artifacts.zip", os.path.join("/tmp", ".A3T")))



from setuptools import find_packages, setup

setup(
    name='a3t',
    packages=find_packages(),
    version='0.0.1',
    description='Implementation for A3T',
    author="Yuhao Zhang",
    license="MIT",
    test_suite="tests",
    install_requires=["forbiddenfruit", "torch>=1.5", "tensorflow>=2.0", "tensorflow-datasets", "nltk>=3.4"],
    # build_requires=["absl-py<0.11,>=0.9"],
    author_email="yuhaoz@cs.wisc.edu",
    long_description="A3T is an adversarial training technique that combines the strengths of augmentation and abstraction techniques. The key idea underlying A3T is to decompose the perturbation space into two subsets, one that can be explored using augmentation and one that can be abstracted.",
    url="https://github.com/ForeverZyh/A3T",
    project_urls={
        "Bug Tracker": "https://github.com/ForeverZyh/A3T/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.5",
)
